"""
App configuration.
"""

import os
import yaml
from platformdirs import PlatformDirs


class AppConfig:
    """
    Configuration wrapper around yaml file.
    """

    def __init__(self, config_file_path):
        """
        Load the required configuration file.
        """
        self._load_config(config_file_path)

    def _load_config(self, config_file_path):
        """
        Load and validate the config file.
        """
        if not os.path.exists(config_file_path):
            raise FileNotFoundError(f"Config file not found at {config_file_path}")

        with open(config_file_path, "r", encoding="utf-8") as file:
            config = yaml.safe_load(file)

        if not isinstance(config, dict):
            raise ValueError("Invalid configuration format. Expected a dictionary.")

        if "DBNAME" not in config:
            raise ValueError("Config file must have 'DBNAME'")

        if "ENV" not in config:
            raise ValueError("Config file must have 'ENV'")

        env = config.get("ENV")
        if env not in ["prod", "dev"]:
            raise ValueError(f"Invalid ENV {env}, can only be prod or dev.")
        self._env = env

        self._mecab_path = None
        if "MECAB_PATH" in config:
            self._mecab_path = config.get("MECAB_PATH")

        self._backup_path = None
        if "BACKUP_PATH" in config:
            self._backup_path = config.get("BACKUP_PATH")

        self._is_docker = False
        if "IS_DOCKER" in config:
            self._is_docker = True

        self._db_name = config.get("DBNAME")
        self._data_path = config.get("DATAPATH", None)
        if self._data_path is None:
            self._data_path = self._get_appdata_dir()

        self._port = config.get("PORT", 5000)

        return config

    def _get_appdata_dir(self):
        "Get user's appdata directory from platformdirs."
        dirs = PlatformDirs("Lute3", "Lute3")
        return dirs.user_data_dir

    @property
    def env(self):
        "App environment (dev or prod)."
        return self._env

    @property
    def is_docker(self):
        "True if the config file has IS_DOCKER."
        return self._is_docker

    @property
    def mecab_path(self):
        "To be used for MECAB_PATH."
        return self._mecab_path

    @property
    def backup_path(self):
        "To be used for BACKUP_PATH."
        return self._backup_path

    @property
    def datapath(self):
        """
        Path to user data / app data.  If not present in the
        dictionary, falls back to platformdirs user_data_dir.
        """
        return self._data_path

    @property
    def userimagespath(self):
        "Path to user images."
        return os.path.join(self.datapath, "userimages")

    @property
    def dbname(self):
        "Database name."
        return self._db_name

    @property
    def is_test_db(self):
        """
        True if the db name starts with test_.

        This is useful because various invoke tasks
        in /invoke.py check if the database is a test_
        db prior to running some destructive action.
        """
        return self.dbname.startswith("test_")

    @property
    def dbfilename(self):
        "Full database file name and path."
        return os.path.join(self.datapath, self.dbname)

    @property
    def sqliteconnstring(self):
        "Full sqlite connection string."
        return f"sqlite:///{self.dbfilename}"

    @property
    def port(self):
        "Port served on."
        return self._port

    @staticmethod
    def create_from_config():
        "Create an AppConfig from the config file."
        thisdir = AppConfig.configdir()
        configfile = os.path.join(thisdir, "config.yml")
        return AppConfig(configfile)

    @staticmethod
    def configdir():
        "Return the path to the configuration file directory."
        return os.path.dirname(os.path.realpath(__file__))



"""
/term routes.
"""

from flask import Blueprint, request, jsonify, render_template, redirect
from lute.models.language import Language
from lute.utils.data_tables import DataTablesFlaskParamParser
from lute.term.datatables import get_data_tables_list
from lute.term.model import Repository, Term
from lute.db import db
from lute.term.forms import TermForm
import lute.utils.formutils

bp = Blueprint("term", __name__, url_prefix="/term")


@bp.route("/index", defaults={"search": None}, methods=["GET"])
@bp.route("/index/<search>", methods=["GET"])
def index(search):
    "Index page."
    return render_template("term/index.html", initial_search=search)


@bp.route("/datatables", methods=["POST"])
def datatables_active_source():
    "Datatables data for terms."
    parameters = DataTablesFlaskParamParser.parse_params(request.form)

    # The DataTablesFlaskParamParser doesn't know about term-specific filters,
    # add those manually.
    filter_param_names = [
        "filtParentsOnly",
        "filtAgeMin",
        "filtAgeMax",
        "filtStatusMin",
        "filtStatusMax",
        "filtIncludeIgnored",
    ]
    request_params = request.form.to_dict(flat=True)
    for p in filter_param_names:
        parameters[p] = request_params.get(p)

    data = get_data_tables_list(parameters)
    return jsonify(data)


def handle_term_form(
    term, repo, form_template_name, return_on_success, embedded_in_reading_frame=False
):
    """
    Handle a form post.

    This is used here and in read.routes -- read.routes.term_form
    lives in an iframe in the reading frames and returns a different
    template on success.
    """
    form = TermForm(obj=term)

    # Flash messages get added on things like term imports.
    # The user opening the form is treated as an acknowledgement.
    term.flash_message = None

    form.language_id.choices = lute.utils.formutils.language_choices()

    if form.validate_on_submit():
        form.populate_obj(term)
        repo.add(term)
        repo.commit()
        return return_on_success

    return render_template(
        form_template_name,
        form=form,
        term=term,
        language_dicts=Language.all_dictionaries(),
        tags=repo.get_term_tags(),
        embedded_in_reading_frame=embedded_in_reading_frame,
    )


def _handle_form(term, repo):
    """
    Handle the form post.  Only show lang. selector
    for new terms.
    """
    return handle_term_form(
        term, repo, "/term/formframes.html", redirect("/term/index", 302)
    )


@bp.route("/edit/<int:termid>", methods=["GET", "POST"])
def edit(termid):
    """
    Edit a term.
    """
    repo = Repository(db)
    term = repo.load(termid)
    return _handle_form(term, repo)


@bp.route("/editbytext/<int:langid>/<text>", methods=["GET", "POST"])
def edit_by_text(langid, text):
    """
    Edit a term.
    """
    repo = Repository(db)
    term = repo.find(langid, text)
    return _handle_form(term, repo)


@bp.route("/new", methods=["GET", "POST"])
def new():
    """
    Create a term.
    """
    repo = Repository(db)
    term = Term()
    return _handle_form(term, repo)


@bp.route("/search/<text>/<int:langid>", methods=["GET"])
def search_by_text_in_language(text, langid):
    "JSON data for parent data."
    repo = Repository(db)
    matches = repo.find_matches(langid, text)
    result = []
    for t in matches:
        result.append({"id": t.id, "text": t.text, "translation": t.translation})
    return jsonify(result)


@bp.route("/sentences/<int:langid>/<text>", methods=["GET"])
def sentences(langid, text):
    "Get sentences for terms."
    repo = Repository(db)
    t = repo.find_or_new(langid, text)
    references = repo.find_references(t)
    return render_template("/term/sentences.html", references=references)


@bp.route("/bulk_update_status", methods=["POST"])
def bulk_update_status():
    "Update the statuses."
    data = request.get_json()
    terms = data.get("terms")
    language_id = int(data.get("langid"))
    new_status = int(data.get("new_status"))

    repo = Repository(db)
    for t in terms:
        term = repo.find_or_new(language_id, t)
        term.status = new_status
        repo.add(term)
    repo.commit()
    return jsonify("ok")


@bp.route("/delete/<int:termid>", methods=["POST"])
def delete(termid):
    """
    Delete a term.
    """
    repo = Repository(db)
    term = repo.load(termid)
    repo.delete(term)
    repo.commit()
    return redirect("/term/index", 302)



"""
Show terms in datatables.
"""

from lute.db import db
from lute.utils.data_tables import DataTablesSqliteQuery, supported_parser_type_criteria


def get_data_tables_list(parameters):
    "Term json data for datatables."

    base_sql = """SELECT
    0 as chk, w.WoID as WoID, LgName, L.LgID as LgID, w.WoText as WoText, parents.parentlist as ParentText, w.WoTranslation,
    replace(wi.WiSource, '.jpeg', '') as WiSource,
    ifnull(tags.taglist, '') as TagList,
    StText,
    StID
    FROM
    words w
    INNER JOIN languages L on L.LgID = w.WoLgID
    INNER JOIN statuses S on S.StID = w.WoStatus
    LEFT OUTER JOIN (
      SELECT WpWoID as WoID, GROUP_CONCAT(PText, ', ') AS parentlist
      FROM
      (
        select WpWoID, WoText as PText
        from wordparents wp
        INNER JOIN words on WoID = WpParentWoID
        order by WoText
      ) parentssrc
      GROUP BY WpWoID
    ) AS parents on parents.WoID = w.WoID
    LEFT OUTER JOIN (
      SELECT WtWoID as WoID, GROUP_CONCAT(TgText, ', ') AS taglist
      FROM
      (
        select WtWoID, TgText
        from wordtags wt
        INNER JOIN tags t on t.TgID = wt.WtTgID
        order by TgText
      ) tagssrc
      GROUP BY WtWoID
    ) AS tags on tags.WoID = w.WoID
    LEFT OUTER JOIN wordimages wi on wi.WiWoID = w.WoID
    """

    filt_parents_only = parameters["filtParentsOnly"]
    filt_age_min = parameters["filtAgeMin"].strip()
    filt_age_max = parameters["filtAgeMax"].strip()
    filt_status_min = int(parameters["filtStatusMin"])
    filt_status_max = int(parameters["filtStatusMax"])
    filt_include_ignored = parameters["filtIncludeIgnored"]

    typecrit = supported_parser_type_criteria()
    wheres = [f"L.LgParserType in ({typecrit})"]
    if filt_parents_only == "true":
        wheres.append("parents.parentlist IS NULL")
    if filt_age_min:
        filt_age_min = int(filt_age_min)
        wheres.append(
            f"cast(julianday('now') - julianday(w.wocreated) as int) >= {filt_age_min}"
        )
    if filt_age_max:
        filt_age_max = int(filt_age_max)
        wheres.append(
            f"cast(julianday('now') - julianday(w.wocreated) as int) <= {filt_age_max}"
        )

    status_wheres = ["StID <> 98"]
    if filt_status_min > 0:
        status_wheres.append(f"StID >= {filt_status_min}")
    if filt_status_max > 0:
        status_wheres.append(f"StID <= {filt_status_max}")

    status_wheres = " AND ".join(status_wheres)
    if filt_include_ignored == "true":
        status_wheres = f"({status_wheres} OR StID = 98)"
    wheres.append(status_wheres)

    where = " AND ".join(wheres)
    full_base_sql = base_sql + " WHERE " + where

    session = db.session
    connection = session.connection()

    return DataTablesSqliteQuery.get_data(full_base_sql, parameters, connection)



"""
Db and image backup.
"""

import os
import shutil
import gzip
from datetime import datetime
import time

from lute.models.setting import SystemSetting


class BackupException(Exception):
    """
    Raised if export bombs for some reason.
    """


def create_backup(app_config, settings, is_manual=False, suffix=None):
    """
    Create backup using current app config, settings.

    is_manual is True if this is a user-triggered manual
    backup, otherwise is False.

    suffix can be specified for test.

    settings are from Setting.get_backup_settings().
      - backup_enabled
      - backup_dir
      - backup_auto
      - backup_warn
      - backup_count
      - last_backup_datetime
    """
    if not os.path.exists(settings.backup_dir):
        raise BackupException("Missing directory " + settings.backup_dir)

    _mirror_images_dir(app_config.userimagespath, settings.backup_dir)

    prefix = "manual_" if is_manual else ""
    if suffix is None:
        suffix = datetime.now().strftime("%Y-%m-%d_%H%M%S")
    fname = f"{prefix}lute_backup_{suffix}.db"
    backupfile = os.path.join(settings.backup_dir, fname)

    f = _create_db_backup(app_config.dbfilename, backupfile)
    _remove_excess_backups(settings.backup_count, settings.backup_dir)
    return f


def should_run_auto_backup(backup_settings):
    """
    True (if applicable) if last backup was old.
    """
    if backup_settings.backup_enabled != "y" or not backup_settings.backup_auto:
        return False

    last = backup_settings.last_backup_datetime
    if last is None:
        return True

    curr = int(time.time())
    diff = curr - last
    return diff > 24 * 60 * 60


def backup_warning(backup_settings):
    "Get warning if needed."
    if not backup_settings.backup_warn:
        return ""

    last = backup_settings.last_backup_datetime
    if last is None:
        return "Never backed up."

    curr = int(time.time())
    diff = curr - last
    old_backup_msg = "Last backup was more than 1 week ago."
    if diff > 7 * 24 * 60 * 60:
        return old_backup_msg

    return ""


def _create_db_backup(dbfilename, backupfile):
    "Make a backup."
    shutil.copy(dbfilename, backupfile)
    f = f"{backupfile}.gz"
    with open(backupfile, "rb") as in_file, gzip.open(
        f, "wb", compresslevel=9
    ) as out_file:
        shutil.copyfileobj(in_file, out_file)
    os.remove(backupfile)
    SystemSetting.set_last_backup_datetime(int(time.time()))
    return f


def _remove_excess_backups(count, outdir):
    "Remove old backups."
    files = [f for f in os.listdir(outdir) if f.startswith("lute_backup_")]
    files.sort(reverse=True)
    to_remove = files[count:]
    for f in to_remove:
        os.remove(os.path.join(outdir, f))


def _mirror_images_dir(userimagespath, outdir):
    "Copy the images to backup."
    target_dir = os.path.join(outdir, "userimages_backup")
    target_dir = os.path.abspath(target_dir)
    if not os.path.exists(target_dir):
        os.mkdir(target_dir)
    shutil.copytree(userimagespath, target_dir, dirs_exist_ok=True)



"""
Lute settings, in settings key-value table.
"""

import datetime
from lute.db import db
from lute.parse.mecab_parser import JapaneseParser
from lute.config.app_config import AppConfig


class SettingBase(db.Model):
    """
    Main settings table.

    This class should not be used, it's the polymorphic base
    for UserSettings and SystemSettings.
    """

    __tablename__ = "settings"
    key = db.Column("StKey", db.String(40), primary_key=True)
    keytype = db.Column("StKeyType", db.String(10), nullable=False)
    value = db.Column("StValue", db.String, nullable=False)
    __mapper_args__ = {"polymorphic_on": keytype}

    @classmethod
    def key_exists_precheck(cls, keyname):
        """
        Check key validity for certain actions.
        """

    @classmethod
    def set_value(cls, keyname, keyvalue):
        "Set, but don't save, a setting."
        cls.key_exists_precheck(keyname)
        s = db.session.query(cls).filter(cls.key == keyname).first()
        if s is None:
            s = cls()
            s.key = keyname
        s.value = keyvalue
        db.session.add(s)

    @classmethod
    def key_exists(cls, keyname):
        "True if exists."
        s = db.session.query(cls).filter(cls.key == keyname).first()
        no_key = s is None
        return not no_key

    @classmethod
    def get_value(cls, keyname):
        "Get the saved key, or None if it doesn't exist."
        cls.key_exists_precheck(keyname)
        s = db.session.query(cls).filter(cls.key == keyname).first()
        if s is None:
            return None
        return s.value

    @classmethod
    def delete_key(cls, keyname):
        "Delete a key."
        s = db.session.query(cls).filter(cls.key == keyname).first()
        if s is not None:
            db.session.delete(s)


class MissingUserSettingKeyException(Exception):
    """
    Cannot set or get unknown user keys.
    """


class UserSetting(SettingBase):
    "User setting."
    __tablename__ = None
    __mapper_args__ = {"polymorphic_identity": "user"}

    @classmethod
    def key_exists_precheck(cls, keyname):
        """
        User keys must exist.
        """
        if not UserSetting.key_exists(keyname):
            raise MissingUserSettingKeyException(keyname)

    @staticmethod
    def _load_settings_from_config():
        """
        Special settings hardcoded in the config file should
        override any settings.

        This is necessary for GitHub CI and Docker, where
        the values should be hardcoded.
        """

        def _set_key(k, v):
            "Set the UserSetting and save."
            if v is not None and v != "":
                UserSetting.set_value(k, v)

        ac = AppConfig.create_from_config()
        _set_key("mecab_path", ac.mecab_path)
        _set_key("backup_dir", ac.backup_path)
        db.session.commit()

    @staticmethod
    def load():
        "Load missing user settings and default values."
        keys_and_defaults = {
            "backup_enabled": None,
            "backup_auto": True,
            "backup_warn": True,
            "backup_dir": None,
            "backup_count": 5,
            "mecab_path": None,
            "custom_styles": "/* Custom css to modify Lute's appearance. */",
        }
        for k, v in keys_and_defaults.items():
            if not UserSetting.key_exists(k):
                s = UserSetting()
                s.key = k
                s.value = v
                db.session.add(s)
        db.session.commit()

        UserSetting._load_settings_from_config()

        # This feels wrong, somehow ... possibly could have an event
        # bus that posts messages about the setting.
        JapaneseParser.set_mecab_path_envkey(UserSetting.get_value("mecab_path"))


class SystemSetting(SettingBase):
    "System setting."
    __tablename__ = None
    __mapper_args__ = {"polymorphic_identity": "system"}

    # Helpers for certain sys settings.

    @classmethod
    def get_last_backup_datetime(cls):
        "Get the last_backup_datetime as int, or None."
        v = cls.get_value("lastbackup")
        if v is None:
            return None
        return int(v)

    @classmethod
    def set_last_backup_datetime(cls, v):
        "Set and save the last backup time."
        cls.set_value("lastbackup", v)
        db.session.commit()


class BackupSettings:
    """
    Convenience wrapper for current backup settings.
    Getter only.
    """

    def __init__(self):
        self.backup_enabled = UserSetting.get_value("backup_enabled")
        self.backup_dir = UserSetting.get_value("backup_dir")

        def _bool(k):
            v = UserSetting.get_value(k)
            return v in (1, "1", "y", True)

        self.backup_auto = _bool("backup_auto")
        self.backup_warn = _bool("backup_warn")
        self.backup_count = int(UserSetting.get_value("backup_count") or 5)
        self.last_backup_datetime = SystemSetting.get_last_backup_datetime()

    @property
    def is_acknowledged(self):
        return self.backup_enabled in ("y", "n")

    @property
    def is_enabled(self):
        return self.backup_enabled == "y"

    @property
    def last_backup_display_date(self):
        "Return the last_backup_datetime as yyyy-mm etc., or None if not set."
        t = self.last_backup_datetime
        if t is None:
            return None
        return datetime.datetime.fromtimestamp(t).strftime("%Y-%m-%d %H:%M:%S")

    @staticmethod
    def get_backup_settings():
        "Get BackupSettings."
        return BackupSettings()



"""
Developer entry point.  Runs lute on a dev server.

You can run with:

inv devstart
python -m devstart

If you want to run this with "python", then for some _extremely odd_
reason, you must run this with the full path to the file.
Ref https://stackoverflow.com/questions/37650208/flask-cant-find-app-file

e.g.:

python /Users/jeff/Documents/Projects/lute_v3/devstart.py
python `pwd`/devstart.py
"""

import os
import logging
from lute.app_factory import create_app
from lute.config.app_config import AppConfig

log = logging.getLogger("werkzeug")
log.setLevel(logging.ERROR)

app_config = AppConfig.create_from_config()

# https://stackoverflow.com/questions/25504149/
#  why-does-running-the-flask-dev-server-run-itself-twice
if os.environ.get("WERKZEUG_RUN_MAIN") != "true":
    # First run
    print()
    print(
        f"Connecting to {app_config.dbname} in folder {app_config.datapath}", flush=True
    )
    print()
else:
    # Reloading.
    pass

app = create_app(app_config, output_func=print)
app.run(debug=True, port=app_config.port)



"""
Settings routes.
"""

import os
from flask import (
    Blueprint,
    request,
    Response,
    render_template,
    redirect,
    flash,
    jsonify,
)
from flask_wtf import FlaskForm
from wtforms import BooleanField, StringField, SelectField, IntegerField, TextAreaField
from wtforms.validators import InputRequired, NumberRange
from wtforms import ValidationError
from lute.config.app_config import AppConfig
from lute.models.language import Language
from lute.models.setting import UserSetting
from lute.db import db
from lute.parse.mecab_parser import JapaneseParser


class UserSettingsForm(FlaskForm):
    """
    Settings.

    Note the field names here must match the keys in the settings table.
    """

    backup_enabled = SelectField(
        "Backup Enabled", choices=[("-", "(not set)"), ("y", "yes"), ("n", "no")]
    )
    backup_dir = StringField("Backup directory")
    backup_auto = BooleanField("Run backups automatically (daily)")
    backup_warn = BooleanField("Warn if backup hasn't run in a week")
    backup_count = IntegerField(
        "Retain backup count",
        validators=[InputRequired(), NumberRange(min=1)],
        render_kw={
            "title": "Count of zipfiles to retain, oldest files are deleted first"
        },
    )

    custom_styles = TextAreaField("Custom styles")

    mecab_path = StringField("MECAB_PATH environment variable")

    def validate_backup_enabled(self, field):
        "User should acknowledge."
        if field.data == "-":
            raise ValidationError('Please change "Backup enabled" to either yes or no.')

    def validate_backup_dir(self, field):
        "Field must be set if enabled."
        if self.backup_enabled.data != "y":
            return
        v = field.data
        if (v or "").strip() == "":
            raise ValidationError("Backup directory required")

        abspath = os.path.abspath(v)
        if v != abspath:
            msg = f'Backup dir must be absolute path.  Did you mean "{abspath}"?'
            raise ValidationError(msg)
        if not os.path.exists(v):
            raise ValidationError(f'Directory "{v}" does not exist.')
        if not os.path.isdir(v):
            raise ValidationError(f'"{v}" is not a directory.')


bp = Blueprint("settings", __name__, url_prefix="/settings")


@bp.route("/index", methods=["GET", "POST"])
def edit_settings():
    "Edit settings."
    form = UserSettingsForm()

    ac = AppConfig.create_from_config()
    if ac.is_docker:
        # User shouldn't change some things with docker.
        kw = {"readonly": True, "style": "background-color: LightGray"}
        form.backup_dir.render_kw = kw
        form.mecab_path.render_kw = kw

    if form.validate_on_submit():
        # Update the settings in the database
        for field in form:
            if field.id not in ("csrf_token", "submit"):
                UserSetting.set_value(field.id, field.data)
        db.session.commit()

        JapaneseParser.set_mecab_path_envkey(form.mecab_path.data)

        flash("Settings updated", "success")
        return redirect("/")

    # Load current settings from the database
    for field in form:
        if field.id != "csrf_token":
            field.data = UserSetting.get_value(field.id)
    # Hack: set boolean settings to ints, otherwise they're always checked.
    form.backup_warn.data = int(form.backup_warn.data or 0)
    form.backup_auto.data = int(form.backup_auto.data or 0)

    return render_template("settings/form.html", form=form)


@bp.route("/test_mecab", methods=["GET"])
def test_parse():
    """
    Do a test parse for the JapaneseParser using the
    given path string.

    Returns { 'success': tokens }, or { 'error' msg }

    """
    mecab_path = request.args.get("mecab_path", None)
    old_key = JapaneseParser.get_mecab_path_envkey()
    result = {"failure": "tbd"}
    try:
        JapaneseParser.set_mecab_path_envkey(mecab_path)
        # Parsing requires a language, even if it's a dummy.
        lang = Language()
        p = JapaneseParser()
        src = "私は元気です"
        toks = p.get_parsed_tokens(src, lang)
        toks = [tok.token for tok in toks if tok.token != "¶"]
        message = f"{src} parsed to [{ ', '.join(toks) }]"
        result = {"result": "success", "message": message}
    except Exception as e:  # pylint: disable=broad-exception-caught
        message = f"{type(e).__name__}: { str(e) }"
        result = {"result": "failure", "message": message}
    finally:
        JapaneseParser.set_mecab_path_envkey(old_key)

    return jsonify(result)


@bp.route("/custom_styles", methods=["GET"])
def custom_styles():
    """
    Return the custom settings for inclusion in the base.html.
    """
    css = UserSetting.get_value("custom_styles")
    response = Response(css, 200)
    response.content_type = "text/css; charset=utf-8"
    return response



"""
User entry point.
"""

import os
import shutil
import logging
from waitress import serve
from lute.app_factory import create_app
from lute.config.app_config import AppConfig

logging.getLogger("waitress.queue").setLevel(logging.ERROR)
logging.getLogger("natto").setLevel(logging.CRITICAL)


def _print(s):
    """
    Print message to stdout.
    """
    if isinstance(s, str):
        s = s.split("\n")
    msg = "\n".join(["  " + lin.strip() for lin in s])
    print(msg, flush=True)


def _create_prod_config_if_needed():
    """
    If config.yml is missing, create one from prod config.
    """
    config_file = os.path.join(AppConfig.configdir(), "config.yml")
    if not os.path.exists(config_file):
        prod_conf = os.path.join(AppConfig.configdir(), "config.yml.prod")
        shutil.copy(prod_conf, config_file)
        _print(["", "Using new production config.", ""])


def start():
    """
    Main entry point: Get the config, init the app, and start.

    If the config file doesn't exist, use the prod config.
    """

    _print(["", "Starting Lute:", ""])

    _create_prod_config_if_needed()

    app_config = AppConfig.create_from_config()

    _print("Initializing app.")
    app = create_app(app_config, output_func=_print)
    _print(f"data path: {app_config.datapath}")
    _print(f"database: {app_config.dbfilename}")
    if app_config.is_docker:
        _print("(Note these are container paths, not host paths.)")

    port = app_config.port
    _print(
        f"""
    Running at:

    http://localhost:{port}
    """
    )

    close_msg = """
    When you're finished reading, stop this process
    with Ctrl-C or your system equivalent.
    """
    if app_config.is_docker:
        close_msg = """
        When you're finished reading, stop this container
        with Ctrl-C, docker compose stop, or docker stop <containerid>
        as appropriate.
        """
    _print(close_msg)

    serve(app, host="0.0.0.0", port=port)


if __name__ == "__main__":
    start()
